# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 11:32:38 2020

@author: Rasmus Zetter
"""


from bfieldtools import line_conductor
from bfieldtools.utils import load_example_mesh


import pytest

import numpy as np


def _fake_line_conductor():
    """
    return synthetic LineConductor
    """
    loops = [
        np.array([[0, 0, 1], [0, 1, 0], [0, 2, 0], [1, 2, 1], [1, 1, 0]]),
        2 * np.array([[0, 0, 1], [2, 1, 0], [1, 2, 0], [1, 2, 1], [1, 1, 0]]),
    ]
    return line_conductor.LineConductor(loops)


def _fake_line_conductor_closed():
    """
    return synthetic LineConductor
    """
    loops = [
        np.array([[0, 0, 1], [0, 1, 0], [0, 2, 0], [1, 2, 1], [1, 1, 0], [0, 0, 1]]),
        2
        * np.array([[0, 0, 1], [2, 1, 0], [1, 2, 0], [1, 2, 1], [1, 1, 0], [0, 0, 1]]),
    ]
    return line_conductor.LineConductor(loops)


def test_line_conductor_functionality():
    """
    Test LineConductor creation and functionality
    """

    lp = _fake_line_conductor()
    lp2 = _fake_line_conductor_closed()

    lp.plot_loops()

    mesh = load_example_mesh("unit_disc")
    scalars = np.ones((len(mesh.vertices),)) - np.linalg.norm(mesh.vertices, axis=1)

    lp = line_conductor.LineConductor(mesh=mesh, scalars=scalars, N_contours=6)

    slp = lp.simplify()

    B = slp.magnetic_field(mesh.vertices + np.array([0, 0, 1]))
    A = slp.vector_potential(mesh.vertices + np.array([0, 0, 1]))
    U = slp.scalar_potential(mesh.vertices + np.array([0, 0, 1]))

    mesh.vertices += np.array([[0, 0, 1]])

    M = slp.mesh_mutual_inductance(mesh)



"""
Computation time for shiftinvert eigenvalue decomp
===========================================================

"""


import numpy as np
from bfieldtools.suhtools import SuhBasis
from mayavi import mlab
import trimesh
import matplotlib.pyplot as plt
from time import clock


# Create basis for a sphere (basis.eigenvals shere the same structure
# as spherical harmonic eigenvalues)
mesh = trimesh.creation.icosphere(4)

closed = False
basis = SuhBasis(mesh, 1)

# Choose Nc and recalculate basis with shift-invert and without
basis.Nc = 100
t0 = clock()
basis.calculate_basis(shiftinvert=True)
print("Time with shift invert:", clock() - t0)
f = mlab.figure()
basis.plot(15, figure=f)
e1 = basis.eigenvals
b1 = basis.basis
t0 = clock()
basis.calculate_basis(shiftinvert=False)
print("Time without shift invert:", clock() - t0)
b2 = basis.basis
f = mlab.figure()
basis.plot(15, figure=f)
e2 = basis.eigenvals

plt.plot(e1)
plt.plot(e2)



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from scipy.linalg import eigh

from bfieldtools.thermal_noise import (
    compute_current_modes,
    visualize_current_modes,
    noise_covar,
    noise_var,
    sensornoise_covar,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling


import pkg_resources

import mne


#%%
font = {"family": "normal", "weight": "normal", "size": 16}

plt.rc("font", **font)

# Fix the simulation parameters
d = 5e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7


Nchunks = 4
quad_degree = 2


#%% Save directory of the MNE example here
SAVE_DIR = ""

#%%

evoked = mne.Evoked(SAVE_DIR + "left_auditory-ave.fif")
evoked.pick_types(meg="mag")

coils = mne.forward._create_meg_coils(evoked.info["chs"], acc=2)

#%% Sensor locations and directions in DEVICE coordinate system

p = np.zeros((len(coils), coils[0]["rmag"].shape[0], 3))
n = np.zeros((len(coils), coils[0]["rmag"].shape[0], 3))
w = np.zeros((len(coils), coils[0]["rmag"].shape[0]))

for i in range(len(coils)):
    p[i] = coils[i]["rmag"]
    n[i] = coils[i]["cosmag"]
    w[i] = coils[i]["w"]


#%%

mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/10x10_plane.obj")
)
mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)

mesh.apply_scale(1e-1)
theta = np.pi / 2


R = np.zeros((4, 4))
R[0] = np.array((np.cos(theta), -1 * np.sin(theta), 0, 0))
R[1] = np.array((np.sin(theta), np.cos(theta), 0, 0))
R[2] = np.array((0, 0, 1, 0))
R[3] = np.array((0, 0, 0, 0))

mesh.apply_transform(R)
mesh.apply_translation((40e-2, 0, 0))


#%%
scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, color=(0, 0, 0), opacity=0.5)

mlab.quiver3d(
    *p.reshape(p.shape[0] * p.shape[1], p.shape[2]).T,
    *n.reshape(n.shape[0] * n.shape[1], n.shape[2]).T,
    mode="arrow"
)
mlab.view(azimuth=90, elevation=90, distance=2.5)

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, color=(0, 0, 0), opacity=0.5)
mlab.quiver3d(*p.T, *n.T, mode="arrow")
mlab.view(azimuth=180, elevation=90, distance=2.5)

#%%

vl, u = compute_current_modes(
    obj=mesh, T=T, resistivity=1 / sigma, thickness=d, mode="AC", return_eigenvals=True
)

noisecov = sensornoise_covar(mesh, p, n, w, vl)
noisecov = noisecov[:, :, 0]

noisecov += noisecov.T - np.diag(np.diag(noisecov))
#%%
plt.figure(figsize=(7.5, 7.5))
plt.imshow(noisecov * 1e30, cmap="RdBu_r")
cbar = plt.colorbar()
plt.clim((-300, 300))
plt.xlabel("Sensor index")
plt.ylabel("Sensor index")
cbar.set_label(r"Noise covariance (fT$^2$/Hz)")
plt.tight_layout()


evoked1 = evoked.copy()
evoked1.data = np.tile(np.sqrt(np.diag(noisecov)), (evoked.times.shape[0], 1)).T
evoked1.plot_topomap(times=0.080, ch_type="mag", vmin=5, vmax=18)



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_covar_dir,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 18}
plt.rc("font", **font)

# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7


# freqs = np.array((0,100, 1000, 10000))
freqs = np.array((0, 50, 100, 500))

Np = 121

z = 0.1
x = np.linspace(-1.25, 1.25, Np)
fp = np.array((x, np.zeros(x.shape), z * np.ones(x.shape))).T

ind_orig = 60

Nchunks = 4
quad_degree = 2

mesh = trimesh.load(
    pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/unitdisc_extremelyfine.stl"
    )
)

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

vl = compute_current_modes(
    obj=mesh,
    T=T,
    resistivity=1 / sigma,
    thickness=d,
    mode="AC",
    freqs=freqs,
    return_eigenvals=False,
)


Bcov = noise_covar(B_coupling, vl) * (1e30)

Bdircov = noise_covar_dir(B_coupling, vl) * (1e30)


#%%

plt.figure(figsize=(5, 5))
plt.plot(x, np.diag(Bcov[:, :, 0, 0]), linewidth=2, label="$B_x$")
plt.plot(x, np.diag(Bcov[:, :, 1, 0]), linewidth=2, label="$B_y$")
plt.plot(x, np.diag(Bcov[:, :, 2, 0]), linewidth=2, label="$B_z$")
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("x (m)")
plt.ylabel("Noise variance (fT$^2$/Hz)")
plt.tight_layout()

plt.figure(figsize=(5, 5))
plt.plot(x, Bcov[ind_orig, :, 0, 0], linewidth=2, label="$B_x$")
plt.plot(x, Bcov[ind_orig, :, 1, 0], linewidth=2, label="$B_y$")
plt.plot(x, Bcov[ind_orig, :, 2, 0], linewidth=2, label="$B_z$")
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("x (m)")
plt.ylabel("Noise covariance (fT$^2$/Hz)")
plt.tight_layout()

plt.figure(figsize=(5, 5))
plt.plot(x, Bcov[ind_orig, :, 2, 0], linewidth=2, label="DC")
plt.plot(x, Bcov[ind_orig, :, 2, 1], linewidth=2, label="50 Hz")
plt.plot(x, Bcov[ind_orig, :, 2, 2], linewidth=2, label="100 Hz")
plt.plot(x, Bcov[ind_orig, :, 2, 3], linewidth=2, label="500 Hz")
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("x (m)")
plt.ylabel(r"$B_z$ covariance (fT$^2$/Hz)")
plt.tight_layout()


plt.figure(figsize=(3, 3))
plt.plot(x, Bcov[ind_orig, :, 2, 0] / np.max(Bcov[ind_orig, :, 2, 0]), linewidth=2)
plt.plot(x, Bcov[ind_orig, :, 2, 1] / np.max(Bcov[ind_orig, :, 2, 1]), linewidth=2)
plt.plot(x, Bcov[ind_orig, :, 2, 2] / np.max(Bcov[ind_orig, :, 2, 2]), linewidth=2)
plt.plot(x, Bcov[ind_orig, :, 2, 3] / np.max(Bcov[ind_orig, :, 2, 3]), linewidth=2)
plt.gca().tick_params(labelbottom=False, labelleft=False)
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.axis("off")
plt.tight_layout()


plt.figure(figsize=(5, 5))
plt.plot(x, Bdircov[:, 0, 1, 0], linewidth=2, label=r"$B_x$$-$$B_y$")
plt.plot(x, Bdircov[:, 0, 2, 0], linewidth=2, label=r"$B_x$$-$$B_z$")
plt.plot(x, Bdircov[:, 1, 2, 0], linewidth=2, label=r"$B_y$$-$$B_z$")
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("x (m)")
plt.ylabel("Noise covariance (fT$^2$/Hz)")
plt.tight_layout()



"""
MAMBA coil
==========

Compact example of a biplanar coil producing homogeneous field in a number of target
regions arranged in a grid. Meant to demonstrate the flexibility in target choice, inspired by the 
technique "multiple-acquisition micro B(0) array" (MAMBA) technique, see https://doi.org/10.1002/mrm.10464

"""


import numpy as np
from mayavi import mlab
import trimesh


from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.contour import scalar_contour
from bfieldtools.viz import plot_3d_current_loops

from bfieldtools.utils import combine_meshes, load_example_mesh


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")

# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 1.5, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))

# Create mesh class object
coil = MeshConductor(mesh_obj=joined_planes, fix_normals=True, basis_name="inner")

#%%
# Set up target and stray field points. Here, the target points are on a planar
# 4x4 grid slightly smaller than the coil dimensions.

center = np.array([0, 0, 0])

sidelength = 0.5
n = 4

height = 0.1
n_height = 2
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-height / 2, height / 2, n_height)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T


grid_target_points = list()
target_field = list()

hori_offsets = [-3, -1, 1, 3]
vert_offsets = [-3, -1, 1, 3]

for i, offset_x in enumerate(hori_offsets):
    for j, offset_y in enumerate(vert_offsets):
        grid_target_points.append(target_points + np.array([offset_x, 0, offset_y]))
        target_field.append((i + j - 3) * np.ones((len(target_points),)))

target_points = np.asarray(grid_target_points).reshape((-1, 3))
target_field = np.asarray(target_field).reshape((-1,))

target_field = np.array(
    [np.zeros((len(target_field),)), target_field, np.zeros((len(target_field),))]
).T


target_abs_error = np.zeros_like(target_field)
target_abs_error[:, 1] += 0.1
target_abs_error[:, 0::2] += 0.1

#%%
# Plot target points and mesh
coil.plot_mesh(opacity=0.1)
mlab.quiver3d(*target_points.T, *target_field.T)


#%%
# Compute coupling matrix that is used to compute the generated magnetic field, create field specification


target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": target_abs_error,
    "target": target_field,
}

#%%
# Run QP solver, plot result

import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)


coil.s.plot()

coil.s.discretize(N_contours=10).plot_loops()



"""
Biplanar coil design
====================

First example in the paper, showing a basic biplanar coil producing homogeneous field in a target
region between the two coil planes.

"""
PLOT = True
SAVE_FIGURES = False

SAVE_DIR = "./Biplanar coil/"


import numpy as np
from mayavi import mlab
import trimesh

from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.contour import scalar_contour
from bfieldtools.viz import plot_3d_current_loops
from bfieldtools.utils import load_example_mesh, combine_meshes


# Set unit, e.g. meter or millimeter.
# This doesn't matter, the problem is scale-invariant
scaling_factor = 1


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")


# Specify coil plane geometry
center_offset = np.array([0, 0, 0]) * scaling_factor
standoff = np.array([0, 5, 0]) * scaling_factor

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))

joined_planes = joined_planes.subdivide()

# Create mesh class object
coil = MeshConductor(
    verts=joined_planes.vertices,
    tris=joined_planes.faces,
    fix_normals=True,
    basis_name="suh",
    N_suh=100,
)

#%%
# Set up target and stray field points

# Here, the target points are on a volumetric grid within a sphere

center = np.array([0, 0, 0]) * scaling_factor

sidelength = 2 * scaling_factor
n = 8
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


#    #Here, the stray field points are on a spherical surface
stray_radius = 20 * scaling_factor

stray_points_mesh = trimesh.creation.icosphere(subdivisions=3, radius=stray_radius)
stray_points = stray_points_mesh.vertices + center

n_stray_points = len(stray_points)


#%%
# Plot geometry
if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

    coil.plot_mesh(representation="wireframe", opacity=0.1, color=(0, 0, 0), figure=f)
    coil.plot_mesh(representation="surface", opacity=0.1, color=(0, 0, 0), figure=f)
    mlab.points3d(*target_points.T, color=(0, 0, 1))
    mlab.points3d(*stray_points.T, scale_factor=0.3, color=(1, 0, 0))

    f.scene.isometric_view()
    f.scene.camera.zoom(1.5)

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "biplanar_geometry.png",
            figure=f,
            magnification=4,
        )
        mlab.close()


#%%
# Create bfield specifications used when optimizing the coil geometry

# The absolute target field amplitude is not of importance,
# and it is scaled to match the C matrix in the optimization function

target_field = np.zeros(target_points.shape)
target_field[:, 0] += 1  # Homogeneous field on X-axis


target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": 0.005,
    "target": target_field,
}
stray_spec = {
    "coupling": coil.B_coupling(stray_points),
    "abs_error": 0.01,
    "target": np.zeros((n_stray_points, 3)),
}

bfield_specification = [target_spec, stray_spec]

#%%
# Run QP solver
import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec, stray_spec],
    objective=(0, 1),
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot coil windings and target points


N_contours = 6

loops = scalar_contour(coil.mesh, coil.s.vert, N_contours=N_contours)

if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(650, 750))
    mlab.clf()

    plot_3d_current_loops(loops, colors="auto", figure=f)

    # B_target = coil.B_coupling(target_points) @ coil.s

    # mlab.quiver3d(*target_points.T, *B_target.T, mode="arrow", scale_factor=1)

    f.scene.isometric_view()
    #    f.scene.camera.zoom(0.95)
    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "biplanar_loops.png",
            figure=f,
            magnification=4,
        )

        mlab.close()


#%%
# Plot continuous stream function

if PLOT:
    from bfieldtools.viz import plot_data_on_vertices

    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
    mlab.clf()

    plot_data_on_vertices(coil.mesh, coil.s.vert, figure=f, ncolors=256)

    f.scene.camera.parallel_projection = 1
    mlab.view(90, 90)
    f.scene.camera.zoom(1.5)

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "biplanar_streamfunction.png",
            figure=f,
            magnification=4,
        )

        mlab.close()



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 16}
plt.rc("font", **font)

# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7
freqs = np.array((0,))


Nchunks = 8
quad_degree = 2


mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/closed_cylinder.stl")
)
mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)


vl = compute_current_modes(
    obj=mesh, T=T, resistivity=1 / sigma, thickness=d, mode="AC", return_eigenvals=False
)


vl[:, 0] = np.zeros(vl[:, 0].shape)  # fix DC-component

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

visualize_current_modes(mesh, vl[:, :, 0], 8, 1)


scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces)
scene.scene.z_minus_view()
surface = scene.children[0].children[0].children[0].children[0]
surface.actor.property.representation = "wireframe"
surface.actor.mapper.scalar_visibility = False
scene.scene.isometric_view()
scene.scene.render()


Np = 30

x = np.linspace(-0.95, 0.95, Np)
fp = np.array((x, np.zeros(x.shape), np.zeros(x.shape))).T

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)
B = noise_var(B_coupling, vl)

a = 0.5
L = 2
rat = L / (2 * a)
Gfact = (
    1
    / (8 * np.pi)
    * (
        (3 * rat**5 + 5 * rat**3 + 2) / (rat**2 * (1 + rat**2) ** 2)
        + 3 * np.arctan(rat)
    )
)
Ban = np.sqrt(Gfact) * mu0 * np.sqrt(kB * T * sigma * d) / a

plt.figure(figsize=(5, 5))
plt.plot(x, Ban * np.ones(x.shape) * 1e15, label="Analytic", linewidth=2)
plt.plot(
    x,
    np.sqrt(B[:, 0]) * 1e15,
    "x",
    label="Numerical",
    markersize=10,
    markeredgewidth=2,
)
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("z (m)")
plt.ylabel(r"$B_z$ noise at DC (fT/rHz)")
plt.tight_layout()

plt.figure()
plt.semilogy(x, np.sqrt(B[:, 0]), label="x")
plt.semilogy(x, np.sqrt(B[:, 1]), label="y")
plt.semilogy(x, np.sqrt(B[:, 2]), "--", label="z")
plt.legend()
plt.xlabel("z (m)")
plt.ylabel("DC noise (T/rHz)")



import numpy as np
import os
from bfieldtools import kicad_io
from bfieldtools.utils import load_example_mesh

from .test_mesh_conductor import _fake_mesh_conductor, _fake_streamfunction

import pytest


def test_kicad_io():

    s = _fake_streamfunction()
    c = s.mesh_conductor

    lp = s.discretize(6)

    kicad_io.python_to_kicad(
        [loop.discrete(lp.vertices) for loop in lp.entities],
        "test_file",
        plane_axes=(0, 1),
        origin=(-1, -1),
        layer="F.Cu",
        net=1,
        scaling=1,
    )

    os.remove("test_file")



"""
Validate calculation bfield calculation from line segments
=================================================================
"""

import numpy as np
import matplotlib.pyplot as plt
import time as t
from bfieldtools.line_magnetics import magnetic_field

""" Bfield calculation from circular current loops using elliptic integrals
"""


def double_factorial(n):
    if n <= 0:
        return 1
    else:
        return n * double_factorial(n - 2)


def bfield_iterative(x, y, z, x_c, y_c, z_c, r, I, n):
    """Compute b field of a current loop using an iterative method to estimate
    elliptic integrals.

    Parameters:
        x, y, z: Evaluation points in 3D space. Accepts matrices and integers.
        x_c, y_c, z_c: Coordinates of the center point of the current loop.
        r: Radius of the current loop.
        I: Current of the current loop.
        n: Number of terms in the serie expansion.

    Returns:
        bfiels (N_points, 3) at evaluation points.

    This calculation is based on paper by Robert A. Schill, Jr (General
    Relation for the Vector Magnetic Field of aCircular Current Loop:
    A Closer Look). DOI: 10.1109/TMAG.2003.808597
    """

    st2 = t.time()
    np.seterr(divide="ignore", invalid="ignore")
    u0 = 4 * np.pi * 1e-7
    Y = u0 * I / (2 * np.pi)

    # Change to cylideric coordinates
    rc = np.sqrt(np.power(x - x_c, 2) + np.power(y - y_c, 2))

    # Coefficients for estimating elliptic integrals with nth degree series
    # expansion using Legendre polynomials
    m = 4 * r * rc / (np.power((rc + r), 2) + np.power((z - z_c), 2))
    K = 1
    E = 1
    for i in range(1, n + 1):
        K = K + np.square(
            double_factorial(2 * i - 1) / double_factorial(2 * i)
        ) * np.power(m, i)
        E = E - np.square(
            double_factorial(2 * i - 1) / double_factorial(2 * i)
        ) * np.power(m, i) / (2 * i - 1)

    K = K * np.pi / 2
    E = E * np.pi / 2

    # Calculation of radial and axial components of B-field
    Brc = (
        Y
        * (z - z_c)
        / (rc * np.sqrt(np.power((rc + r), 2) + np.power((z - z_c), 2)))
        * (
            -K
            + E
            * (np.power(rc, 2) + np.power(r, 2) + np.power((z - z_c), 2))
            / (np.power((rc - r), 2) + np.power((z - z_c), 2))
        )
    )

    Bz = (
        Y
        / (np.sqrt(np.power((rc + r), 2) + np.power((z - z_c), 2)))
        * (
            K
            - E
            * (np.power(rc, 2) - np.power(r, 2) + np.power((z - z_c), 2))
            / (np.power((rc - r), 2) + np.power((z - z_c), 2))
        )
    )

    # Set nan and inf values to 0
    Brc[np.isinf(Brc)] = 0
    Bz[np.isnan(Bz)] = 0
    Brc[np.isnan(Brc)] = 0
    Bz[np.isinf(Bz)] = 0

    # Change back to cartesian coordinates
    Bx = Brc * (x - x_c) / rc
    By = Brc * (y - y_c) / rc

    # Change nan values from coordinate transfer to 0
    Bx[np.isnan(Bx)] = 0
    By[np.isnan(By)] = 0

    B = np.zeros((3, X.size), dtype=np.float64)
    B[0] = Bx.flatten()
    B[1] = By.flatten()
    B[2] = Bz.flatten()

    et2 = t.time()
    print("Execution time for iterative method is:", et2 - st2)

    return B.T


""" Plot field of a circular current path
"""
x = np.linspace(-1, 1, 100)
Ntheta = 10000
theta = np.linspace(0, 2 * np.pi, Ntheta)
vertices = np.zeros((Ntheta, 3), dtype=np.float64)
vertices[:, 0] = np.cos(theta) * 0.1
vertices[:, 1] = np.sin(theta) * 0.1
vertices[:, 2] = 0.2

X, Y = np.meshgrid(x, x, indexing="ij")
Z = np.zeros((x.size, x.size))

points = np.zeros((3, X.size), dtype=np.float64)
points[0] = X.flatten()
points[1] = Y.flatten()

b1 = magnetic_field(vertices, points.T)  # Calculates discretised bfield
b2 = bfield_iterative(X, Y, Z, 0, 0, 0.2, 0.1, 1, 25)  # Calculates bfield iteratively

# Error between two calculation methods.
berr = (b2 - b1) / b1 * 100
BE = berr.T[2]  # By changing the index, errors in different components can be obtained
ind = np.where(np.abs(BE) > 0.1)  # The limit for significant error is set to 0.1%
bpoints = points.T[ind]

from mayavi import mlab

mlab.figure(1)
q = mlab.quiver3d(*points, *b1.T)
q.glyph.glyph_source.glyph_position = "center"
mlab.plot3d(*vertices.T)

mlab.figure(2)
q = mlab.quiver3d(*points, *b2.T)
q.glyph.glyph_source.glyph_position = "center"
mlab.plot3d(*vertices.T)

plt.figure(3)
plt.hist(berr.T[2], bins=50, density=True, histtype="bar")
plt.title("Histogram of error between calculation methods.")
plt.xlabel("%")

#%% Plot the b-field vectors exceeding the error limit
if len(bpoints > 0):

    from mayavi import mlab

    mlab.figure(3)
    q = mlab.quiver3d(*bpoints.T, *b1[ind].T)
    q.glyph.glyph_source.glyph_position = "center"
    mlab.plot3d(*vertices.T)

    q = mlab.quiver3d(*bpoints.T, *b2[ind].T)
    q.glyph.glyph_source.glyph_position = "center"
